-- ###########################################################################
--
--  Iriel's attempt at some addon utilities to make life easier when
--  creating addons.
--
-- ###########################################################################
-- Version 2
--   * 2005-09-12 -- Improved cleanup if found to not be most recent version.
--   * 2005-08-14 -- Separated initializer and actual tables so that its
--                   impossible to avoid GetInstance.
--                   Added CheckInstance()
--                   Fixed bug with flag arguments being false
-- Version 1
--   * 2005-08-12 -- Initial version

-- Initialize the basic initializer object structure
local utilInit = {
   -- version number
   ['version'] = 2;
};

-- Initialize the basic object structure
local util = {
   -- Initializer
   ['init'] = utilInit;

   -- The fallback message catalog with some basic messages
   ['FALLBACK_MESSAGE_CATALOG'] = {
      ["CMD_AVAILABLE_HEADER"] = 'Available commands:',
      ["CMD_AVAILABLE"] = '  |cffffff88%s%s|r - %s',
      ["CMD_AVAILABLE_LIST"] = 'Available commands: |cffffff88%s|r',
      ["CMD_AVAILABLE_ALSOLIST"] = 'Other commands: |cffffff88%s|r',
      ["ERR_UNKNOWN_COMMAND"] = "Unknown command: '%s'",
      ["ERR_COMMAND_NOHANDLER"] = "Unknown argument type %s for command '%s'",
      ["MISSING_MESSAGE"] = "MISSING MESSAGE CATALOG ENTRY - '%s'",
   }
};

-- This craziness is used so that I dont have dangling local symbols in this
-- file in order to ensure that unused copies of the utilities dont hang
-- around in memory.
local function createGetInstance()
   -- Metatable for utility objects
   local utilMetatable = {
      ["__index"] = util;
   }

   -- Create a new utility instance
   -- catalog -- Message catalog
   function utilInit:GetInstance(catalog)
      if (not catalog) then
	 catalog = {};
      end
      local newInstance = {
	 ["MESSAGE_CATALOG"] = catalog
      }
      setmetatable(newInstance, utilMetatable);
      return newInstance;
   end
end
createGetInstance();
createGetInstance = nil;

-- Create a new utility instance with settings copied from another
-- orig -- The instance to copy from
function utilInit:CopyInstance(orig)
   local newInstance = {
      ["MESSAGE_CATALOG"] = orig.MESSAGE_CATALOG or {};
   }
   setmetatable(newInstance, utilMetatable);
   return newInstance;
end

-- Check this instance against the current version and replace it if
-- there's a more recent version available.
--
-- Returns:
--   The utility object to use (self if no changes, otherwise a new instance)
function util:CheckInstance()
   if (self.init == IrielAddonUtils) then
      return self;
   end
   return IrielAddonUtils:CopyInstance(self);
end

function util:LookupMessage(key, useFallback)
   local msg = self.MESSAGE_CATALOG[key];
   if (useFallback and (not msg)) then
      msg = self.FALLBACK_MESSAGE_CATALOG[key];
   end
   return msg;
end

function util:GetMessage(key, ...)
   local msg = self.MESSAGE_CATALOG[key];
   if (msg == nil) then
      msg = self.FALLBACK_MESSAGE_CATALOG[key];
   end

   if (msg == nil) then
      if (key == 'MISSING_MESSAGE') then
	 message("Cannot find message catalog entry " .. key);
      end
      local missing = self:GetMessage("MISSING_MESSAGE", key);
      return missing;
   elseif (arg.n > 0) then
      return string.format(msg, unpack(arg));
   else
      return msg;
   end
end

function util:Write(key, ...)
   if (DEFAULT_CHAT_FRAME) then
      local msg = self:GetMessage(key, unpack(arg));
      DEFAULT_CHAT_FRAME:AddMessage(msg);
   end
end

function util:WriteRaw(msg)
   if (DEFAULT_CHAT_FRAME) then
      DEFAULT_CHAT_FRAME:AddMessage(msg);
   end
end

function util:CheckRangeArg(label, low, high, argstr)
   local arg = tonumber(argstr);
   if (not arg) then
      self:Write("ERR_" .. label .. "_BADNUMBER", argstr, low, high);
      return nil
   end
   if ((arg < low) or (arg > high)) then
      self:Write("ERR_" .. label .. "_BADRANGE", arg, low, high);
      return nil
   end
   return arg;
end

function util:CheckNumberArg(label, argstr)
   local arg = tonumber(argstr);
   if (not arg) then
      self:Write("ERR_" .. label .. "_BADNUMBER", argstr, low, high);
      return nil
   end
   return arg;
end

function util:CheckFlagArg(label, argstr)
   local arg = string.lower(argstr);
   if ((arg == "yes") 
       or (arg == "on") 
	  or (arg == "1") 
	  or (arg == "true")) then
      return true;
   end
   if ((arg == "no") 
       or (arg == "off") 
	  or (arg == "0") 
	  or (arg == "false")) then
      return false;
   end
   self:Write("ERR_" .. label .. "_BADFLAG", argstr,
	      "yes, no, on, off, true, false");
   return nil
end

-- Command structure
--
--  label     Description of command
--  msgkey  (Optional) Message catalog label for errors
--  arglabel  (Optional) Description of argument
--  action    Action function to call when processed
--  argtype   (Default - 'ACTION') Which command handler to use
--            Can be ACTION, RANGE, FLAG
--
--  +RANGE command args
--    rangelo   Low end of range
--    rangehi   High end of range

local handlers = {};
util.COMMAND_HANDLERS = handlers;

function handlers.ACTION(util, command, rest)
   command.action(rest, util);
end

function handlers.RANGE(util, command, rest)
   local val = util:CheckRangeArg(command.msgkey, 
				  command.rangelo, command.rangehi, rest);
   if (val) then
      command.action(val, util);
   end
end

function handlers.FLAG(util, command, rest)
   local val = util:CheckFlagArg(command.msgkey, rest);
   if (val ~= nil) then
      command.action(val, util);
   end
end

function handlers.NUMBER(util, command, rest)
   local val = util:CheckNumberArg(command.msgkey, rest);
   if (val) then
      command.action(val, util);
   end
end

-- Throw away for GC
handlers = nil;

-- :ParseCommand
--   commands -- Command table
--   line -- Command line
function util:ParseCommand(commands, line)
   local b,e,command,rest = string.find(line, "^%s*([^%s]+)%s*(.*)$");

   if (not b) then
      local defaultCommand = commands[""];
      if (defaultCommand and defaultCommand.action) then
	 if (defaultCommand.action(self)) then
	    return;
	 end
      end

      local sortedCommands = {};
      for cmd, commandData in pairs(commands) do
	 if (cmd ~= '') then
	    table.insert(sortedCommands, cmd);
	 end
      end
      table.sort(sortedCommands);

      local nolabel = {};
      local anylabel = false;

      for idx, cmd in ipairs(sortedCommands) do
	 local commandData = commands[cmd];

	 local key = commandData.msgkey;
	 local label = self:LookupMessage("CMD_" .. key .. "_LABEL");
	 local arglabel = self:LookupMessage("CMD_" .. key .. "_ARGLABEL");
	 if (not label) then
	    if (arglabel) then
	       table.insert(nolabel .. ' ' .. arglabel, cmd);
	    else
	       table.insert(nolabel, cmd);
	    end
	 else
	    if (not anylabel) then
	       anylabel = true;
	       local msg = self:LookupMessage("CMD_AVAILABLE_HEADER", true);
	       if (msg and (msg ~= '')) then
		  self:WriteRaw(msg);
	       end
	    end
	    if (arglabel) then
	       self:Write("CMD_AVAILABLE", cmd, ' ' .. arglabel, label);
	    else
	       self:Write("CMD_AVAILABLE", cmd, '', label);
	    end
	 end
      end

      if (table.getn(nolabel) > 0) then
	 if (anylabel) then
	    self:Write("CMD_AVAILABLE_ALSOLIST", table.concat(nolabel, ", "));
	 else
	    self:Write("CMD_AVAILABLE_LIST", table.concat(nolabel, ", "));
	 end
      end
      return;
   end

   local commandData = commands[string.lower(command)];

   if (commandData) then
      local handler =
	 self.COMMAND_HANDLERS[(commandData.argtype or 'ACTION')];
      if (not handler) then
	 self:Write("ERR_COMMAND_NOHANDLER", handler, command);
	 return;
      end
      handler(self, commandData, rest);
      return;
   end

   self:Write("ERR_UNKNOWN_COMMAND", command);
end

-- Update the global reference if it doesn't exist, or if this is a more
-- recent version.
if ((not IrielAddonUtils) or (IrielAddonUtils.version < utilInit.verson)) then
   IrielAddonUtils = utilInit;
else
   -- Throw away this copy
   util = nil;
   utilInit = nil;
end
